
%if 0

Test hook program
 2021 by C. Masloch

Usage of the works is permitted provided that this
instrument is retained with the works, so that any entity
that uses the works is notified of this instrument.

DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY.

%endif

%include "lmacros3.mac"
%include "AMIS.MAC"

_uninstall	equ 1
_current	equ 2
_nop		equ 4
_ihpfs		equ 8
_reference	equ 100h
_gotnumber	equ 200h


	cpu 8086
	org 256
start:
	jmp strict short setup

iispentry intr, 0, intr, 2
	jmp far [cs:.next]

iisphwreset intr

residentsizep: equ ($ + 15 - $$ + 256) >> 4


setup:
	mov dl, 'K'		; default signature letter
	mov bx, 21h		; default interrupt number
	mov si, 81h
	lframe none
	lenter
	lvar word, flags
	xor ax, ax
	 push ax
.loop:
@@:
	lodsb
	cmp al, 9
	je @B
	cmp al, 32
	jb end
	je @B
	cmp al, '/'
	je .switch
	cmp al, '-'
	je .switch
	mov ah, al
	call capitalise
	call getnybble
	jc .letter
	xor bx, bx
	mov bl, al
	lodsb
	call getnybble
	jc .check_end_number
	add bx, bx
	add bx, bx
	add bx, bx
	add bx, bx
	add bl, al
	lodsb
.check_end_number:
	cmp al, 9
	je @F
	cmp al, 32
	ja .invalid
@@:
	testopt [bp + ?flags], _gotnumber
	jnz .toomany
	setopt [bp + ?flags], _gotnumber
	dec si
	jmp .loop

.toomany:
	mov dx, toomanymessage
	jmp abort_msg

.letter:
	cmp al, 'X'
	jne .invalid
.loop_letter:
	mov dl, ah
	jmp .loop

.switch:
	lodsb
	call capitalise
	cmp al, 'L'
	je .switch_l
	cmp al, 'U'
	je .switch_u
	cmp al, 'C'
	je .switch_c
	cmp al, 'N'
	je .switch_n
	cmp al, 'I'
	je .switch_i
	cmp al, 'H'
	je .switch_help
	cmp al, '?'
	je .switch_help
	jmp .invalid

.switch_l:
@@:
	lodsb
	cmp al, ':'
	je @B
	cmp al, '='
	je @B
	mov ah, al
	call capitalise
	cmp al, 'A'
	jb .invalid
	cmp al, 'Z'
	ja .invalid
	jmp .loop_letter

.switch_u:
	setopt [bp + ?flags], _uninstall
	jmp .loop

.switch_c:
	setopt [bp + ?flags], _current
	jmp .loop

.switch_n:
	setopt [bp + ?flags], _nop
	jmp .loop

.switch_i:
	setopt [bp + ?flags], _ihpfs
	jmp .loop

.switch_help:
	mov dx, helpmessage
	jmp exit_msg

.invalid:
	mov dx, errormessage
abort_msg:
	mov ah, 09h
	int 21h
	mov ax, 4CFFh
	int 21h

errormessage:
	ascicline "Invalid paramter."

toomanymessage:
	ascicline "Too many interrupt numbers."

helpmessage:
	db "TESTHOOK - Install test hooks not discoverable via AMIS",13,10
	db "Free software by C. Masloch",13,10
	db 13,10
	db "Parameters:",13,10
	db "        NN      Operate on interrupt number NNh",13,10
	db "        x       (Literal 'x') Alias for switch /L=x",13,10
	db "Switches:",13,10
	db "        /N      Insert NOPs before ieStart",13,10
	db "        /I      Make iHPFS style uninstalled header",13,10
	db "        /L=x    Patch letter x into ieSignature",13,10
	db "        /C      Update current installed handler (must be reachable)",13,10
	db "        /U      Uninstall handler (must be reachable)",13,10
	db 36


end:
	lvar word, interruptnumber
	 push bx
	lvar word, letter
	 push dx
	lvar dword, reference
	 push ax
	 push ax

	testopt [bp + ?flags], _uninstall | _current
	jz install

findinstalled:
	mov ah, 35h
	mov al, bl
	int 21h
.loop:
	cmp bx, intr
	jne .next
	cmp byte [es:bx + ieSignature + 1], 'B'
	jne .next
	push es
	mov di, es
	dec di
	mov es, di
	mov di, 8
	mov si, mcbname
	mov cx, di
	repe cmpsb
	pop es
	je .found
.next:
	call IsIISPEntry?			; that an IISP entry ?
	jnz .notfound				; nope --> (NZ)
	setopt [bp + ?flags], _reference
	mov word [bp + ?reference], bx
	mov word [bp + ?reference + 2], es
	les bx, [es:bx + ieNext ]		; get next interrupt entry
	jmp .loop

.notfound:
	mov dx, notfoundmessage
	jmp abort_msg

notfoundmessage: equ $
	ascicline "Not found."

.found:
	testopt [bp + ?flags], _uninstall
	jnz uninstall

current:
	mov byte [es:bx + ieSignature], dl
	mov di, 0EA90h
	testopt [bp + ?flags], _ihpfs
	jnz @F
	mov di, 10EBh
	testopt [bp + ?flags], _nop
	jz @F
	mov word [es:bx + ieStart - 2], 9090h
	mov di, 0EEBh
@@:
	mov word [es:bx + ieEntry], di
	mov dx, currentmessage
exit_msg:
	mov ah, 09h
	int 21h
	mov ax, 4C00h
	int 21h

currentmessage: equ $
	ascicline "Made current."

uninstall:
	testopt [bp + ?flags], _reference
	jz @F
	push es				; preserve resident segment
	push es
	pop ds				; ds:si -> our resident handler
	mov si, bx
	les di, [bp + ?reference]	; es:di -> IISP entry referencing ours
	call update_iisp_header		; update IISP header
	pop es				; = resident segment
	jmp .common

@@:
	lds dx, [es:bx + ieNext]	; -> next
	mov ah, 25h
	int 21h				; write next handler to IVT
.common:
	mov ah, 49h
	int 21h				; free our resident memory
	push cs
	pop ds
	mov dx, uninstallmessage
	jmp exit_msg

uninstallmessage: equ $
	ascicline "Uninstalled."


install:
	mov ax, cs
	dec ax
	mov es, ax
	mov di, 8
	mov si, mcbname
	mov cx, di
	rep movsb
	mov byte [intr + ieSignature], dl
	testopt [bp + ?flags], _ihpfs
	jz @F
	mov word [intr + ieEntry], 0EA90h
	jmp @FF

@@:
	testopt [bp + ?flags], _nop
	jz @F
	mov word [intr + ieStart - 2], 9090h
	mov word [intr + ieEntry], 0EEBh
@@:
	mov ah, 35h
	mov al, bl
	int 21h
	mov word [intr + ieNext], bx
	mov word [intr + ieNext + 2], es
	mov dx, intr
	mov ah, 25h
	int 21h
	xor ax, ax
	xchg ax, word [2Ch]
	mov es, ax
	mov ah, 49h
	int 21h
	xor bx, bx
	mov cx, word [32h]
@@:
	mov ah, 3Eh
	int 21h
	inc bx
	loop @B
	mov ax, 3100h
	mov dx, residentsizep
	int 21h


	align 2
mcbname:
	fill 8, 32, db "TESTHOOK"


capitalise:
	cmp al, 'a'
	jb .notlowercase
	cmp al, 'z'
	ja .notlowercase
	and al, ~ 20h		; Uppercase if required
.notlowercase:
	retn


getnybble:
	cmp al, '0'
	jb .invalid		; unknown character -->
	cmp al, '9'
	jbe @F
	call capitalise
	cmp al, 'A'
	jb .invalid
	cmp al, 'F'
	ja .invalid
	sub al, 'A' - 10 - '0'
@@:
	sub al, '0'
	db __TEST_IMM8		; (NC, skip stc)
.invalid:
	stc
	retn


		; INP:	es:bx-> interrupt entry
		; OUT:	NZ if non-IISP entry
		;	ZR if IISP entry
IsIISPEntry?:
	cmp bx, - (ieSignature + 2)		; may access word at offset FFFFh ?
	ja .return				; yes, avoid --> (NZ)
	cmp word [ es:bx + ieSignature ], "KB"	; "KB"/424Bh ? ("BK" in MASM)
	jne .return
	cmp word [ es:bx + ieEntry ], 0EA90h	; nop\jmp far imm16:imm16 ?
	je .return				; unused IISP entry (created by iHPFS) -->
	cmp byte [ es:bx + ieEntry ], 0EBh	; jmp short ... ?
		; (This opcode should strictly be jmp short $+18 but there's programs
		; that save an additional jmp opcode by jumping directly into their
		; code even though it's not right behind the header.)
	jne .return
	cmp byte [ es:bx + ieJmphwreset ], 0EBh	; jmp short ... ?
	je .return				; usual IISP entry -->
	cmp byte [ es:bx + ieJmphwreset ], 0CBh	; retf ?
	je .return				; a shorter variant -->
	cmp byte [ es:bx + ieJmphwreset ], 0CFh	; iret ?
.return:
	retn


%assign _PM 0

	align 2, db 0
debuggeramissig:
.ven:	fill 8,32,db "ecm"	; vendor
.prod:	fill 8,32,db "lDebug"	; product


		; INP:	ds:si -> source IISP header (or pseudo header)
		;	es:di -> destination IISP header
		; OUT:	EI
		;	si and di both incremented by 6
		; CHG:	-
		; STT:	UP
update_iisp_header:
	push ax
	push cx
	push dx
	push di
	push si
	push es
%if _PM
	 call selector_to_segment
%endif
	push ds
	 push ss
	 pop ds

findinstalleddebugger:
%if 0
	mov al, 2Dh
	push bx
	call intchk
	pop bx
	jz .zero

	testopt [options3], opt3_no_call_update
	jnz .zero

	mov ah, byte [try_debugger_amis_multiplex_number]
	call .check
	jnc @F
%endif

	mov ah, 0FFh		; start with multiplex number 0FFh
.loop:
	call .check
@@:
	mov al, 30h		; al = 30h to indicate found, ah = multiplex number
	jnc .end
	sub ah, 1		; search is backward (to find latest installed first), from 0FFh to 00h including
	jnc .loop		; try next if we didn't check all yet -->

.zero:
	xor ax, ax		; al = 0 to indicate none found
	jmp .end		;  If not found, continue -->


		; INP:	ah = multiplex number to check
		;	ds = ss = cs
		; OUT:	CY if multiplex number unused or no signature match,
		;	 bp, ah, ds unmodified
		;	NC if match found,
		;	 ah = multiplex number (unmodified)
		; CHG:	si, di, es, cx, dx
.check:
%if 0
	testopt [internalflags4], dif4_int_2D_hooked
	jz @F
	cmp ah, byte [amis_multiplex_number]
	je .notfound		; do not use our own multiplexer -->
%endif
@@:
	mov al, 00h		; AMIS installation check
%if _PM
	call call_int2D
%else
	int 2Dh			; AMIS (or "DOS reserved" = iret if no AMIS present)
%endif
	cmp al, 0FFh
	jne .notfound
	mov si, debuggeramissig	; ds:si -> our AMIS name strings
%if _PM
	call setes2dx
%else
	mov es, dx		; es:di -> name strings of AMIS multiplexer that just answered
%endif
	mov cx, 8		; Ignore description, only compare vendor and program name
	repe cmpsw
	je .checkret		; ZR, NC = match -->
.notfound:
	stc			; NZ, CY no match
.checkret:
	retn

.end:
	; mov word [debuggerfunction], ax

	pop ds			; must be not using scratchsel !
%if _PM
	pop dx
	call setes2dx		; this one uses scratchsel
%else
	pop es
%endif
	pop si
	pop di
	pop dx
	pop cx
	; push ax
	; mov ax, word [ss:debuggerfunction]
	test ax, ax			; found the debugger ?
	jz @F				; no -->
%if _PM
	call ispm
	jnz .86m
[cpu 286]
	push es				; es
	 call selector_to_segment	; convert to segment
	push ds				; ds
	 call selector_to_segment	; convert to segment
	push word 2Dh			; int 2Dh
	push bp				; bp
	call intcall_return_parameter_es_parameter_ds
	add sp, 4			; discard returned ds, es
__CPU__
	db __TEST_IMM16			; (skip int)
%endif
.86m:
	int 2Dh				; call its Update IISP Header function
	cmp al, 0FFh			; supported ?
	pop ax
	je .ret				; yes. done -->
	db __TEST_IMM8			; (skip pop)
@@:
	pop ax				; restore ax, then do manual update
%if _PM
   push ax
  call push_if
%else
  pushf
%endif
	cli				; try to rest while updating chain
	cmpsw				; skip over first word (entrypoint)
					;  (generally xxEBh or 0EA90h)
	movsw
	movsw				; transfer source ieNext to dest ieNext
%if _PM
  call pop_if
%else
  popf
%endif
.ret:
	retn
